"""User table creation

Revision ID: 3a45dbc8708e
Revises:
Create Date: 2019-01-19 18:22:07.129516

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3a45dbc8708e'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('users',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('username', sa.String(length=50), nullable=False),
    sa.Column('email', sa.String(length=120), nullable=False),
    sa.Column('password', sa.String(length=255), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('email'),
    sa.UniqueConstraint('username')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('users')
    # ### end Alembic commands ###



from flask import Blueprint, jsonify, request

from .. import db
from ..users.utils import authenticate
from .model import Category

categories_blueprint = Blueprint('categories', __name__)


@categories_blueprint.route('/categories', methods=['GET'])
@authenticate
def get_user_categories(user_id):
    categories = Category.query.filter_by(id=user_id).all()
    response_object = {
        'status': 'success',
        'data': [category.serialize() for category in categories],
    }
    return jsonify(response_object), 200


@categories_blueprint.route('/categories', methods=['POST'])
@authenticate
def add_user_category(user_id):
    post_data = request.get_json()
    if not post_data or post_data.get('name') is None:
        response_object = {'status': 'error', 'message': 'Invalid payload.'}
        return jsonify(response_object), 400

    name = post_data.get('name').lower()
    category = Category.query.filter_by(user_id=user_id, name=name).first()
    if category:
        response_object = {
            'status': 'error',
            'message': f'A category named "{name}" already exists.',
        }
        return jsonify(response_object), 400

    new_category = Category(user_id=user_id, name=name)
    if post_data.get('description'):
        new_category.description = post_data.get('description')
    db.session.add(new_category)
    db.session.commit()
    response_object = {'status': 'success', 'data': [new_category.serialize()]}
    return jsonify(response_object), 201


@categories_blueprint.route('/categories/<int:cat_id>', methods=['PATCH'])
@authenticate
def update_user_category(user_id, cat_id):
    post_data = request.get_json()
    if not post_data:
        response_object = {'status': 'error', 'message': 'Invalid payload.'}
        return jsonify(response_object), 400

    category = Category.query.filter_by(id=cat_id).first()
    if not category or category.user_id != user_id:
        response_object = {
            'status': 'not found',
            'message': f'Category not found.',
        }
        return jsonify(response_object), 404
    name = post_data.get('name').lower()
    if (
        name
        and Category.query.filter(
            Category.user_id == user_id,
            Category.name == name,
            Category.id != cat_id,
        ).first()
    ):
        response_object = {
            'status': 'error',
            'message': f'A category named "{name}" already exists.',
        }
        return jsonify(response_object), 400
    if name:
        category.name = post_data.get('name')
    if post_data.get('description'):
        category.description = post_data.get('description')
    db.session.commit()
    response_object = {'status': 'success', 'data': [category.serialize()]}
    return jsonify(response_object), 200


@categories_blueprint.route('/categories/<int:cat_id>', methods=['DELETE'])
@authenticate
def delete_user_category(user_id, cat_id):
    category = Category.query.filter_by(id=cat_id).first()
    if not category or category.user_id != user_id:
        response_object = {
            'status': 'not found',
            'message': f'Category not found.',
        }
        return jsonify(response_object), 404
    if category.is_default:
        response_object = {
            'status': 'error',
            'message': 'Default category can not be deleted.',
        }
        return jsonify(response_object), 400
    if category.articles:
        # default category exists, no need to check it
        default_category = Category.query.filter_by(
            user_id=user_id, is_default=True
        ).first()
        for article in category.articles:
            article.category_id = default_category.id
            db.session.commit()
    db.session.delete(category)
    db.session.commit()
    response_object = {'status': 'no content'}
    return jsonify(response_object), 204



# source: http://docs.gunicorn.org/en/stable/custom.html
from __future__ import unicode_literals

import multiprocessing
import os

import gunicorn.app.base
from flask_migrate import upgrade
from gunicorn.six import iteritems
from rdltr import create_app


def number_of_workers():
    return (multiprocessing.cpu_count() * 2) + 1


os.environ['RDLTR_SETTINGS'] = 'rdltr.config.ProductionConfig'
HOST = os.getenv('RDLTR_HOST', 'localhost')
PORT = os.getenv('RDLTR_PORT', '5000')
WORKERS = os.getenv('RDLTR_WORKERS', number_of_workers())
BASEDIR = os.path.abspath(os.path.dirname(__file__))
app = create_app()


class StandaloneApplication(gunicorn.app.base.BaseApplication):
    def __init__(self, app, options=None):
        self.options = options or {}
        self.application = app
        super(StandaloneApplication, self).__init__()

    def load_config(self):
        config = dict(
            [
                (key, value)
                for key, value in iteritems(self.options)
                if key in self.cfg.settings and value is not None
            ]
        )
        for key, value in iteritems(config):
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application


def upgrade_db():
    with app.app_context():
        upgrade(directory=BASEDIR + '/migrations')


def main():
    options = {'bind': f"{HOST}:{PORT}", 'workers': WORKERS}
    StandaloneApplication(app, options).run()


if __name__ == '__main__':
    main()



import os


class BaseConfig:
    """Base configuration"""

    DEBUG = False
    TESTING = False
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    BCRYPT_LOG_ROUNDS = 13
    TOKEN_EXPIRATION_DAYS = 30
    TOKEN_EXPIRATION_SECONDS = 0
    REGISTRATION_ALLOWED = (
        False if os.getenv('RDLTR_ALLOW_REGISTRATION') == "false" else True
    )


class DevelopmentConfig(BaseConfig):
    """Development configuration"""

    DEBUG = True
    SQLALCHEMY_DATABASE_URI = os.getenv('RDLTR_DB_URL', 'sqlite:///rdltr.db')
    SECRET_KEY = 'development key'
    USERNAME = 'admin'
    PASSWORD = 'default'
    BCRYPT_LOG_ROUNDS = 4


class TestingConfig(BaseConfig):
    """Testing configuration"""

    DEBUG = True
    TESTING = True
    SQLALCHEMY_DATABASE_URI = os.getenv(
        'RDLTR_DB_TEST_URL', 'sqlite:///rdltr_test.db'
    )
    SECRET_KEY = 'test key'
    USERNAME = 'admin'
    PASSWORD = 'default'
    BCRYPT_LOG_ROUNDS = 4
    TOKEN_EXPIRATION_DAYS = 0
    TOKEN_EXPIRATION_SECONDS = 3


class ProductionConfig(BaseConfig):
    """Production configuration"""

    DEBUG = False
    SQLALCHEMY_DATABASE_URI = os.getenv('RDLTR_DB_URL')
    SECRET_KEY = os.getenv('RDLTR_SECRET_KEY')



from flask import Blueprint, jsonify, request

from .. import db
from ..users.utils import authenticate
from .model import Tag

tags_blueprint = Blueprint('tags', __name__)


@tags_blueprint.route('/tags', methods=['GET'])
@authenticate
def get_user_tags(user_id):
    tags = Tag.query.filter_by(id=user_id).all()
    response_object = {
        'status': 'success',
        'data': [tag.serialize() for tag in tags],
    }
    return jsonify(response_object), 200


@tags_blueprint.route('/tags', methods=['POST'])
@authenticate
def add_user_tag(user_id):
    post_data = request.get_json()
    if not post_data or post_data.get('name') is None:
        response_object = {'status': 'error', 'message': 'Invalid payload.'}
        return jsonify(response_object), 400

    name = post_data.get('name')
    tag = Tag.query.filter_by(user_id=user_id, name=name).first()
    if tag:
        response_object = {
            'status': 'error',
            'message': f'A tag named "{name}" already exists.',
        }
        return jsonify(response_object), 400

    new_tag = Tag(user_id=user_id, name=name)
    db.session.add(new_tag)
    db.session.commit()
    response_object = {'status': 'success', 'data': [new_tag.serialize()]}
    return jsonify(response_object), 201


@tags_blueprint.route('/tags/<int:tag_id>', methods=['PATCH'])
@authenticate
def update_user_tag(user_id, tag_id):
    post_data = request.get_json()
    if not post_data:
        response_object = {'status': 'error', 'message': 'Invalid payload.'}
        return jsonify(response_object), 400

    tag = Tag.query.filter_by(id=tag_id).first()
    if not tag or tag.user_id != user_id:
        response_object = {'status': 'not found', 'message': 'Tag no found.'}
        return jsonify(response_object), 404
    if post_data.get('name'):
        tag.name = post_data.get('name')
    db.session.commit()
    response_object = {'status': 'success', 'data': [tag.serialize()]}
    return jsonify(response_object), 200


@tags_blueprint.route('/tags/<int:tag_id>', methods=['DELETE'])
@authenticate
def delete_user_tag(user_id, tag_id):
    tag = Tag.query.filter_by(id=tag_id).first()
    if not tag or tag.user_id != user_id:
        response_object = {'status': 'not found', 'message': 'Tag no found.'}
        return jsonify(response_object), 404
    db.session.delete(tag)
    db.session.commit()
    response_object = {'status': 'no content'}
    return jsonify(response_object), 204



import datetime

from sqlalchemy import UniqueConstraint

from .. import db
from ..users.model import User

tags_to_articles = db.Table(
    'articles_tags',
    db.Column(
        'tag_id', db.Integer, db.ForeignKey('tags.id'), primary_key=True
    ),
    db.Column(
        'page_id', db.Integer, db.ForeignKey('articles.id'), primary_key=True
    ),
)


class Category(db.Model):
    __tablename__ = "categories"
    __table_args__ = (
        UniqueConstraint('user_id', 'name', name='category_unique_name'),
    )
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    user_id = db.Column(
        db.Integer,
        db.ForeignKey('users.id', name='fk_user_category'),
        nullable=True,
    )
    name = db.Column(db.String(50), nullable=False)
    description = db.Column(db.String(200))
    is_default = db.Column(db.Boolean, default=False)
    user = db.relationship(User, backref='category_user')
    articles = db.relationship('Article', backref='category_articles')

    def __init__(self, user_id, name):
        self.user_id = user_id
        self.name = name

    def serialize(self):
        return {
            'id': self.id,
            'user_id': self.user_id,
            'name': self.name,
            'description': self.description,
            'is_default': self.is_default,
        }


class Tag(db.Model):
    __tablename__ = "tags"
    __table_args__ = (
        UniqueConstraint('user_id', 'name', name='tag_unique_name'),
    )
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    user_id = db.Column(
        db.Integer,
        db.ForeignKey('users.id', name='fk_user_tags'),
        nullable=True,
    )
    name = db.Column(db.String(50), nullable=False)
    user = db.relationship(User, backref='tag_user')

    def __init__(self, user_id, name):
        self.user_id = user_id
        self.name = name

    def serialize(self):
        return {'id': self.id, 'user_id': self.user_id, 'name': self.name}


class Article(db.Model):
    __tablename__ = "articles"
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    category_id = db.Column(
        db.Integer,
        db.ForeignKey('categories.id', name='fk_category_articles'),
        nullable=True,
    )
    url = db.Column(db.String(), nullable=False)
    title = db.Column(db.String(), nullable=False)
    content = db.Column(db.String(), nullable=False)
    html_content = db.Column(db.String(), nullable=False)
    date_added = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    comments = db.Column(db.String())
    read_status = db.Column(db.Boolean(), default=False, nullable=False)
    category = db.relationship(Category, backref='category')
    tags = db.relationship(
        Tag,
        secondary=tags_to_articles,
        lazy='subquery',
        backref=db.backref('articles', lazy=True),
    )

    def __init__(self, category_id, url, title, content, html_content):
        self.category_id = category_id
        self.url = url
        self.title = title
        self.content = content
        self.html_content = html_content

    def serialize(self):
        return {
            'id': self.id,
            'url': self.url,
            'title': self.title,
            'html_content': self.html_content,
            'category': self.category.serialize(),
            'tags': [tag.serialize() for tag in self.tags],
            'comments': self.comments,
            'date_added': self.date_added,
            'read': self.read_status,
        }



import logging
import os

from flask import Flask, jsonify, render_template
from flask_bcrypt import Bcrypt
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy

log_file = os.getenv('RDLTR_LOG')
logging.basicConfig(
    filename=log_file,
    format='%(asctime)s - %(name)s - %(levelname)s - ' '%(message)s',
    datefmt='%Y/%m/%d %H:%M:%S',
)
app_log = logging.getLogger('rdltr')
bcrypt = Bcrypt()
db = SQLAlchemy()
migrate = Migrate()


def create_app():
    app = Flask(__name__, static_folder="dist/static", template_folder="dist")
    app_log.setLevel(logging.DEBUG if app.debug else logging.INFO)

    # set config
    with app.app_context():
        app_settings = os.getenv(
            'RDLTR_SETTINGS', 'rdltr.config.DevelopmentConfig'
        )
        app.config.from_object(app_settings)

        # set up extensions
        bcrypt.init_app(app)
        db.init_app(app)
        migrate.init_app(app, db, directory='rdltr/migrations')

    if app.debug:
        logging.getLogger('sqlalchemy').setLevel(logging.WARNING)
        logging.getLogger('sqlalchemy').handlers = logging.getLogger(
            'werkzeug'
        ).handlers
        logging.getLogger('sqlalchemy.orm').setLevel(logging.WARNING)
        logging.getLogger('flake8').propagate = False
        app_log.setLevel(logging.DEBUG)

    if app.debug:
        # Enable CORS
        @app.after_request
        def after_request(response):
            response.headers.add('Access-Control-Allow-Origin', '*')
            response.headers.add(
                'Access-Control-Allow-Headers', 'Content-Type,Authorization'
            )
            response.headers.add(
                'Access-Control-Allow-Methods',
                'GET,PUT,POST,DELETE,PATCH,OPTIONS',
            )
            return response

    from .articles.model import Article, Category, Tag  # noqa
    from .users.model import User  # noqa

    from .articles.articles import articles_blueprint  # noqa
    from .articles.categories import categories_blueprint  # noqa
    from .articles.tags import tags_blueprint  # noqa
    from .users.auth import auth_blueprint  # noqa

    app.register_blueprint(articles_blueprint, url_prefix='/api')
    app.register_blueprint(auth_blueprint, url_prefix='/api')
    app.register_blueprint(categories_blueprint, url_prefix='/api')
    app.register_blueprint(tags_blueprint, url_prefix='/api')

    @app.route('/api/ping')
    def ping_pong():
        return jsonify({'status': 'success', 'message': 'pong!'})

    @app.route('/', defaults={'path': ''})
    @app.route('/<path:path>')
    def catch_all(path):
        return render_template('index.html')

    return app



from flask import Blueprint, current_app, jsonify, request
from sqlalchemy import exc, or_

from .. import app_log, bcrypt, db
from ..articles.model import Category
from .model import User
from .utils import authenticate, passwords_controls, register_controls

auth_blueprint = Blueprint('auth', __name__)


def handle_error(e, db):
    db.session.rollback()
    app_log.error(e)

    response_object = {
        'status': 'error',
        'message': 'Error. Please try again or contact the administrator.',
    }
    return jsonify(response_object), 500


@auth_blueprint.route('/auth/register', methods=['POST'])
def register_user():
    if not current_app.config.get('REGISTRATION_ALLOWED'):
        response_object = {
            'status': 'error',
            'message': 'Error. Registration is disabled.',
        }
        return jsonify(response_object), 403

    # get post data
    post_data = request.get_json()
    if (
        not post_data
        or post_data.get('username') is None
        or post_data.get('email') is None
        or post_data.get('password') is None
        or post_data.get('password_conf') is None
    ):
        response_object = {'status': 'error', 'message': 'Invalid payload.'}
        return jsonify(response_object), 400
    username = post_data.get('username')
    email = post_data.get('email')
    password = post_data.get('password')
    password_conf = post_data.get('password_conf')

    try:
        ret = register_controls(username, email, password, password_conf)
    except TypeError as e:
        app_log.error(e)

        response_object = {
            'status': 'error',
            'message': 'Error. Please try again or contact the administrator.',
        }
        return jsonify(response_object), 500
    if ret != '':
        response_object = {'status': 'error', 'message': 'Errors: ' + ret}
        return jsonify(response_object), 400

    try:
        # check for existing user
        user = User.query.filter(
            or_(User.username == username, User.email == email)
        ).first()
        if not user:
            # add new user to db
            new_user = User(username=username, email=email, password=password)
            db.session.add(new_user)
            db.session.flush()

            new_category = Category(user_id=new_user.id, name='default')
            new_category.description = 'Default category'
            new_category.is_default = True
            db.session.add(new_category)
            # generate auth token
            auth_token = new_user.encode_auth_token(new_user.id)
            db.session.commit()
            response_object = {
                'status': 'success',
                'message': 'Successfully registered.',
                'auth_token': auth_token.decode(),
                'user': new_user.serialize(),
            }
            return jsonify(response_object), 201
        else:
            response_object = {
                'status': 'error',
                'message': 'Sorry. That user already exists.',
            }
            return jsonify(response_object), 400
    # handler errors
    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error(e, db)


@auth_blueprint.route('/auth/login', methods=['POST'])
def login_user():
    # get post data
    post_data = request.get_json()
    if not post_data:
        response_object = {'status': 'error', 'message': 'Invalid payload.'}
        return jsonify(response_object), 400
    email = post_data.get('email')
    password = post_data.get('password')
    try:
        # check for existing user
        user = User.query.filter(User.email == email).first()
        if user and bcrypt.check_password_hash(user.password, password):
            # generate auth token
            auth_token = user.encode_auth_token(user.id)
            response_object = {
                'status': 'success',
                'message': 'Successfully logged in.',
                'auth_token': auth_token.decode(),
                'user': user.serialize(),
            }
            return jsonify(response_object), 200
        else:
            response_object = {
                'status': 'error',
                'message': 'Invalid credentials.',
            }
            return jsonify(response_object), 400
    # handler errors
    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error(e, db)


@auth_blueprint.route('/auth/logout', methods=['GET'])
@authenticate
def logout_user(user_id):
    # all checks are made by @authenticate
    response_object = {
        'status': 'success',
        'message': 'Successfully logged out.',
    }
    return jsonify(response_object), 200


@auth_blueprint.route('/auth/profile', methods=['GET'])
@authenticate
def get_user_status(user_id):
    user = User.query.filter_by(id=user_id).first()
    response_object = {'status': 'success', 'user': user.serialize()}
    return jsonify(response_object), 200


@auth_blueprint.route('/auth/profile/edit', methods=['POST'])
@authenticate
def update_password(user_id):
    # get post data
    post_data = request.get_json()
    if (
        not post_data
        or post_data.get('old_password') is None
        or post_data.get('new_password') is None
        or post_data.get('new_password_conf') is None
    ):
        response_object = {'status': 'error', 'message': 'Invalid payload.'}
        return jsonify(response_object), 400

    user = User.query.filter_by(id=user_id).first()

    old_password = post_data.get('old_password')
    if not bcrypt.check_password_hash(user.password, old_password):
        response_object = {
            'status': 'error',
            'message': 'Invalid credentials.',
        }
        return jsonify(response_object), 400

    password = post_data.get('new_password')
    password_conf = post_data.get('new_password_conf')
    try:
        ret = passwords_controls(password, password_conf)
        if ret != '':
            response_object = {'status': 'error', 'message': 'Errors: ' + ret}
            return jsonify(response_object), 400

        user.password = bcrypt.generate_password_hash(
            password, current_app.config.get('BCRYPT_LOG_ROUNDS')
        ).decode()
        db.session.commit()

        response_object = {'status': 'success', 'user': user.serialize()}
        return jsonify(response_object), 200
    except (
        exc.IntegrityError,
        exc.OperationalError,
        ValueError,
        TypeError,
    ) as e:
        return handle_error(e, db)



import re
from functools import wraps

from flask import jsonify, request

from .model import User


def is_valid_email(email):
    mail_pattern = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"
    return re.match(mail_pattern, email) is not None


def register_controls(username, email, password, password_conf):
    ret = ''
    if not 2 < len(username) < 13:
        ret += 'Username: 3 to 12 characters required.\n'
    if not is_valid_email(email):
        ret += 'Valid email must be provided.\n'
    ret += passwords_controls(password, password_conf)
    return ret


def passwords_controls(password, password_conf):
    ret = ''
    if password != password_conf:
        ret += 'Password and password confirmation don\'t match.\n'
    if len(password) < 8:
        ret += 'Password: 8 characters required.\n'
    return ret


def verify_user(current_request):
    response_object = {
        'status': 'error',
        'message': 'Something went wrong. Please contact us.',
    }
    code = 401
    auth_header = current_request.headers.get('Authorization')
    if not auth_header:
        response_object['message'] = 'Provide a valid auth token.'
        return response_object, code, None
    auth_token = auth_header.split(" ")[1]
    resp = User.decode_auth_token(auth_token)
    if isinstance(resp, str):
        response_object['message'] = resp
        return response_object, code, None
    user = User.query.filter_by(id=resp).first()
    if not user:
        return response_object, code, None
    return None, None, resp


def authenticate(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        response_object, code, resp = verify_user(request)
        if response_object:
            return jsonify(response_object), code
        return f(resp, *args, **kwargs)

    return decorated_function



import re

import requests
from bs4 import BeautifulSoup
from flask import Blueprint, jsonify, request
from readability import Document
from requests import ConnectionError
from sqlalchemy import exc, or_

from .. import app_log, db
from ..users.utils import authenticate
from .model import Article, Category, Tag

articles_blueprint = Blueprint('articles', __name__)


@articles_blueprint.route('/articles', methods=['GET'])
@authenticate
def get_user_articles(user_id):
    params = request.args.copy()
    tag_id = params.get('tag_id')
    page = 1 if 'page' not in params.keys() else int(params.get('page'))
    category_id = params.get('cat_id')
    query = params.get('q')
    only_not_read = params.get('not_read')
    articles_pagination = (
        Article.query.join(Category)
        .filter(
            or_(
                Article.title.like('%' + query + '%'),
                Article.content.like('%' + query + '%'),
            )
            if query
            else True,
            Article.read_status == False if only_not_read else True,  # noqa
            Category.user_id == user_id,
            Category.id == category_id if category_id else True,
            Article.tags.any(id=tag_id) if tag_id else True,
        )
        .order_by(Article.date_added.desc())
        .paginate(page, 12, False)
    )

    articles = articles_pagination.items
    response_object = {
        'status': 'success',
        'data': [article.serialize() for article in articles],
        'pagination': {
            'has_next': articles_pagination.has_next,
            'has_prev': articles_pagination.has_prev,
            'page': articles_pagination.page,
            'pages': articles_pagination.pages,
            'total': articles_pagination.total,
        },
    }
    return jsonify(response_object), 200


@articles_blueprint.route('/articles/<int:article_id>', methods=['GET'])
@authenticate
def get_user_article(user_id, article_id):
    article = Article.query.filter_by(id=article_id).first()
    if not article or article.category.user_id != user_id:
        response_object = {
            'status': 'not found',
            'message': f'Article not found.',
        }
        return jsonify(response_object), 404

    response_object = {'status': 'success', 'data': [article.serialize()]}
    return jsonify(response_object), 200


@articles_blueprint.route('/articles', methods=['POST'])
@authenticate
def add_user_article(user_id):
    post_data = request.get_json()
    if not post_data or post_data.get('url') is None:
        response_object = {'status': 'error', 'message': 'Invalid payload.'}
        return jsonify(response_object), 400

    url = post_data.get('url').strip()

    # check if url is valid (regex from Django validator)
    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # noqa
        r'localhost|'
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|'
        r'\[?[A-F0-9]*:[A-F0-9:]+\]?)'
        r'(?::\d+)?'
        r'(?:/?|[/?]\S+)$',
        re.IGNORECASE,
    )
    if not re.match(regex, url):
        response_object = {
            'status': 'error',
            'message': 'Error: Invalid URL, please check it.',
        }
        return jsonify(response_object), 400

    try:
        response = requests.get(url)
        doc = Document(response.text)
        title = doc.title()
        # 'html_content' is used for display
        # and existing classes are removed
        html_content = re.sub(
            'class=".*?"', '', doc.summary(html_partial=False)
        )
        # 'content' is used for search
        content = BeautifulSoup(html_content, "html.parser").text
    except ConnectionError as e:
        app_log.error(e)
        response_object = {
            'status': 'error',
            'message': 'Error. Cannot connect to the URL, please check it.',
        }
        return jsonify(response_object), 500
    except Exception as e:
        app_log.error(e)
        response_object = {
            'status': 'error',
            'message': 'Error. Please try again or contact the administrator.',
        }
        return jsonify(response_object), 500

    category_id = post_data.get('category_id')
    if not category_id:
        category = Category.query.filter_by(
            user_id=user_id, is_default=True
        ).first()
    else:
        category = Category.query.filter_by(id=category_id).first()

    if not category:
        response_object = {
            'status': 'error',
            'message': 'Article category not found.',
        }
        return jsonify(response_object), 500

    new_article = Article(
        category_id=category.id,
        title=title,
        content=content,
        html_content=html_content,
        url=url,
    )
    db.session.flush()
    tags = post_data.get('tags')
    if tags and isinstance(tags, list):
        for tag_name in tags:
            tag = Tag.query.filter_by(user_id=user_id, name=tag_name).first()
            if not tag:
                tag = Tag(user_id=user_id, name=tag_name)
                db.session.flush()
            new_article.tags.append(tag)
            db.session.flush()
    db.session.add(new_article)
    db.session.commit()
    response_object = {'status': 'success', 'data': [new_article.serialize()]}
    return jsonify(response_object), 201


@articles_blueprint.route('/articles/<int:article_id>', methods=['PATCH'])
@authenticate
def update_user_category(user_id, article_id):
    post_data = request.get_json()
    if not post_data:
        response_object = {'status': 'error', 'message': 'Invalid payload.'}
        return jsonify(response_object), 400

    article = Article.query.filter_by(id=article_id).first()
    if not article or article.category.user_id != user_id:
        response_object = {
            'status': 'not found',
            'message': f'Article not found.',
        }
        return jsonify(response_object), 404

    category_id = post_data.get('category_id')
    if category_id:
        category = Category.query.filter_by(id=category_id).first()
        if not category or category.user_id != user_id:
            response_object = {
                'status': 'error',
                'message': f'Article category not found.',
            }
            return jsonify(response_object), 500
        article.category_id = category.id
    try:
        comments = post_data.get('comments')
        if comments:
            article.comments = comments
        read_status = post_data.get('update_read_status')
        if read_status is not None:
            article.read_status = read_status
        new_tags = post_data.get('tags')
        if isinstance(new_tags, list):
            if not new_tags:
                article.tags.clear()
            else:
                existing_tags = [tag for tag in article.tags]
                for existing_tag in existing_tags:
                    if existing_tag.name not in new_tags:
                        article.tags.remove(existing_tag)
                        db.session.flush()
                for tag_name in new_tags:
                    tag = Tag.query.filter_by(
                        user_id=user_id, name=tag_name
                    ).first()
                    if not tag:
                        tag = Tag(user_id=user_id, name=tag_name)
                        db.session.flush()
                    article.tags.append(tag)
                    db.session.flush()
        db.session.commit()
        response_object = {'status': 'success', 'data': [article.serialize()]}
        return jsonify(response_object), 200
    except (
        exc.IntegrityError,
        exc.OperationalError,
        exc.InterfaceError,
        exc.StatementError,
    ) as e:
        db.session.rollback()
        app_log.error(e)
        response_object = {
            'status': 'error',
            'message': 'Error. Please try again or contact the administrator.',
        }
        return jsonify(response_object), 500


@articles_blueprint.route('/articles/<int:article_id>', methods=['DELETE'])
@authenticate
def delete_user_category(user_id, article_id):
    article = Article.query.filter_by(id=article_id).first()
    if not article or article.category.user_id != user_id:
        response_object = {
            'status': 'not found',
            'message': f'Article not found.',
        }
        return jsonify(response_object), 404
    db.session.delete(article)
    db.session.commit()
    response_object = {'status': 'no content'}
    return jsonify(response_object), 204



import datetime

import jwt
from flask import current_app

from .. import bcrypt, db


class User(db.Model):
    __tablename__ = "users"
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    username = db.Column(db.String(50), nullable=False, unique=True)
    email = db.Column(db.String(120), nullable=False, unique=True)
    password = db.Column(db.String(255), nullable=False)
    created_at = db.Column(db.DateTime, nullable=False)
    categories = db.relationship('Category', backref='user_categories')
    tags = db.relationship('Tag', backref='user_tags')

    def __repr__(self):
        return f'<User {self.username!r}>'

    def __init__(
        self, username, email, password, created_at=datetime.datetime.utcnow()
    ):
        self.username = username
        self.email = email
        self.created_at = created_at
        self.password = bcrypt.generate_password_hash(
            password, current_app.config.get('BCRYPT_LOG_ROUNDS')
        ).decode()

    @staticmethod
    def encode_auth_token(user_id):
        """
        Generates the auth token
        :param user_id: -
        :return: JWToken
        """
        payload = {
            'exp': datetime.datetime.utcnow()
            + datetime.timedelta(
                days=current_app.config.get('TOKEN_EXPIRATION_DAYS'),
                seconds=current_app.config.get('TOKEN_EXPIRATION_SECONDS'),
            ),
            'iat': datetime.datetime.utcnow(),
            'sub': user_id,
        }
        return jwt.encode(
            payload, current_app.config.get('SECRET_KEY'), algorithm='HS256'
        )

    @staticmethod
    def decode_auth_token(auth_token):
        """
        Decodes the auth token
        :param auth_token: -
        :return: integer|string
        """
        try:
            payload = jwt.decode(
                auth_token,
                current_app.config.get('SECRET_KEY'),
                algorithms=['HS256'],
            )
            return payload['sub']
        except jwt.ExpiredSignatureError:
            return 'Signature expired. Please log in again.'
        except jwt.InvalidTokenError:
            return 'Invalid token. Please log in again.'

    def serialize(self):
        return {
            'id': self.id,
            'username': self.username,
            'email': self.email,
            'created_at': self.created_at,
            'categories': [
                category.serialize() for category in self.categories
            ],
            'tags': [tag.serialize() for tag in self.tags],
        }



"""Article, Category and Tag creation

Revision ID: 214f7fe3b1ee
Revises: 9b52e07d1b4b
Create Date: 2019-02-04 20:03:52.255730

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9b52e07d1b4b'
down_revision = '3a45dbc8708e'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('categories',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('name', sa.String(length=50), nullable=False),
    sa.Column('description', sa.String(length=200), nullable=True),
    sa.Column('is_default', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], name='fk_user_category'),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('user_id', 'name', name='category_unique_name')
    )
    op.create_table('tags',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('name', sa.String(length=50), nullable=False),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], name='fk_user_tags'),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('user_id', 'name', name='tag_unique_name')
    )
    op.create_table('articles',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('category_id', sa.Integer(), nullable=True),
    sa.Column('url', sa.String(), nullable=False),
    sa.Column('title', sa.String(), nullable=False),
    sa.Column('content', sa.String(), nullable=False),
    sa.Column('html_content', sa.String(), nullable=False),
    sa.Column('date_added', sa.DateTime(), nullable=True),
    sa.Column('comments', sa.String(), nullable=True),
    sa.Column('read_status', sa.Boolean(), nullable=False),
    sa.ForeignKeyConstraint(['category_id'], ['categories.id'], name='fk_category_articles'),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('articles_tags',
    sa.Column('tag_id', sa.Integer(), nullable=False),
    sa.Column('page_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['page_id'], ['articles.id'], ),
    sa.ForeignKeyConstraint(['tag_id'], ['tags.id'], ),
    sa.PrimaryKeyConstraint('tag_id', 'page_id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('articles_tags')
    op.drop_table('articles')
    op.drop_table('tags')
    op.drop_table('categories')
    # ### end Alembic commands ###
